/**
 * The mergeable-store module
 * @packageDocumentation
 * @module mergeable-store
 * @since v5.0.0
 */

import {CellOrUndefined, Content, Store, ValueOrUndefined} from './store.d.ts';
import {Id} from './common.d.ts';

/**
 * The Hash type
 * @category Stamps
 */
export type Hash = number;

/**
 * The Time type
 * @category Stamps
 */
export type Time = string;

/**
 * The Stamp type
 * @category Stamps
 */
export type Stamp<Thing, Hashed extends boolean = false> = Hashed extends true
  ? [thing: Thing, time: Time, hash: Hash]
  : [thing: Thing, time?: Time];

// ContentHashes
export type ContentHashes = [tablesHash: Hash, valuesHash: Hash];

// TablesStamp
export type TablesStamp<Hashed extends boolean = false> = Stamp<
  {[tableId: Id]: TableStamp<Hashed>},
  Hashed
>;

// TableHashes
export type TableHashes = {[tableId: Id]: Hash};

// TableStamp
export type TableStamp<Hashed extends boolean = false> = Stamp<
  {[rowId: Id]: RowStamp<Hashed>},
  Hashed
>;

// RowHashes
export type RowHashes = {[tableId: Id]: {[rowId: Id]: Hash}};

// RowStamp
export type RowStamp<Hashed extends boolean = false> = Stamp<
  {[cellId: Id]: CellStamp<Hashed>},
  Hashed
>;

// CellHashes
export type CellHashes = {[tableId: Id]: {[rowId: Id]: {[cellId: Id]: Hash}}};

// CellStamp
export type CellStamp<Hashed extends boolean = false> = Stamp<
  CellOrUndefined,
  Hashed
>;

// ValuesStamp
export type ValuesStamp<Hashed extends boolean = false> = Stamp<
  {[valueId: Id]: ValueStamp<Hashed>},
  Hashed
>;

// ValuesHashes
export type ValuesHashes = {[valueId: Id]: Hash};

// ValueStamp
export type ValueStamp<Hashed extends boolean = false> = Stamp<
  ValueOrUndefined,
  Hashed
>;

/**
 * The MergeableContent type
 * @category Mergeable
 */
export type MergeableContent = [
  mergeableTables: TablesStamp<true>,
  mergeableValues: ValuesStamp<true>,
];

/**
 * The MergeableChanges type
 * @category Mergeable
 */
export type MergeableChanges = [
  mergeableTables: TablesStamp,
  mergeableValues: ValuesStamp,
  isChanges: 1,
];

/**
 * The MergeableStore type
 * @category Mergeable
 */
export interface MergeableStore extends Store {
  //
  /**
   * The getMergeableContent method
   * @category Getter
   */
  getMergeableContent(): MergeableContent;

  /**
   * The getMergeableContentHashes method
   * @category Syncing
   */
  getMergeableContentHashes(): ContentHashes;

  /**
   * The getMergeableTableHashes method
   * @category Syncing
   */
  getMergeableTableHashes(): TableHashes;

  /**
   * The getMergeableTableDiff method
   * @category Syncing
   */
  getMergeableTableDiff(
    relativeTo: TableHashes,
  ): [newTables: TablesStamp, differentTableHashes: TableHashes];

  /**
   * The getMergeableRowHashes method
   * @category Syncing
   */
  getMergeableRowHashes(otherTableHashes: TableHashes): RowHashes;

  /**
   * The getMergeableRowDiff method
   * @category Syncing
   */
  getMergeableRowDiff(
    otherTableRowHashes: RowHashes,
  ): [newRows: TablesStamp, differentRowHashes: RowHashes];

  /**
   * The getMergeableCellHashes method
   * @category Syncing
   */
  getMergeableCellHashes(otherTableRowHashes: RowHashes): CellHashes;

  /**
   * The getMergeableCellDiff method
   * @category Syncing
   */
  getMergeableCellDiff(otherTableRowCellHashes: CellHashes): TablesStamp;

  /**
   * The getMergeableValuesHashes method
   * @category Syncing
   */
  getMergeableValuesHashes(): ValuesHashes;

  /**
   * The getMergeableValueDiff method
   * @category Syncing
   */
  getMergeableValueDiff(relativeTo: ValuesHashes): ValuesStamp;

  /**
   * The setMergeableContent method
   * @category Setter
   */
  setMergeableContent(mergeableContent: MergeableContent): MergeableStore;

  /**
   * The setDefaultContent method
   * @category Setter
   */
  setDefaultContent(content: Content): MergeableStore;

  /**
   * The getTransactionMergeableChanges method
   * @category Getter
   */
  getTransactionMergeableChanges(): MergeableChanges;

  /**
   * The applyMergeableChanges method
   * @category Setter
   */
  applyMergeableChanges(
    mergeableChanges: MergeableChanges | MergeableContent,
  ): MergeableStore;

  /**
   * The merge method
   * @category Setter
   */
  merge(mergeableStore: MergeableStore): MergeableStore;
}

/**
 * The createMergeableStore function
 * @category Creation
 */
export function createMergeableStore(uniqueId?: Id): MergeableStore;
